#Build an MG collection from user's surfed pages.

import os
import sys
import string
import gzip
import tempfile
import traceback

SUCCESS = 1
FAILURE = 0

def writeDebugInfo(msg):
	debugFile = open("c:/debug.txt", "a+")
	debugFile.write(msg)
	debugFile.close()
	return

def tracebackAsString(limit):
	type, value, tb = sys.exc_info()
	aList = traceback.format_exception(type, value, tb, limit)
	formattedString = ""
	for s in aList:
		formattedString = formattedString + s + "\n"
	return formattedString

def	buildTempFile(storageFolder, tempDirectory, userName):
	try:
		tempfile.tempdir = tempDirectory
		tempFileName = tempfile.mktemp()
		tf = open(tempFileName, "w")
		pagesDir = storageFolder + "/" + userName

		surfedPages = os.listdir(pagesDir)
		first = 1
		#for all compressed surfed pages in the user's surfed files directory
		for page in surfedPages:
			if not first:
				tf.write("\n")	#write control-B at end of each file if there will be another file
			#read compressed page
			compressedFileName = pagesDir + "/" + page
			#writeDebugInfo("Compressed page file name=" + compressedFileName + "\n")

			#for some reason we're getting a strange error here about
			#AttributeError: unused_data
			#f = gzip.GzipFile(compressedFileName)

			f = open(compressedFileName)	#not really compressed
			lines = f.readlines()
			f.close()

			#write decompressed page to temp file
			for line in lines:
				tf.write(line)
			first = 0
		tf.close()

		return tempFileName
	except Exception, data:
		writeDebugInfo("Error building temp file for MG collection build: "
						+ tracebackAsString(25) + "\n")
		return None

def buildCollection(userName, tempDirectory, storageFolder, mgDir, mgCollection):
	collectionsDirectory = mgDir + userName + "/"
	collectionBaseName = mgCollection

	try:
		os.chdir(collectionsDirectory)
	except Exception, data:
		s = "Error changing to collections dir for MG collection build: " + `data` + "\n"
		writeDebugInfo(s)
		return FAILURE

	# Build a temp file from all surfed pages, adding control-B's in-between.
	inputFiles = buildTempFile(storageFolder, tempDirectory, userName)
	if inputFiles == None:
		return FAILURE

	# Create *.text.stats, *.invf.dict.build,
	# *.invf.chunk and *.invf.chunks.trans
	try:
		os.system("mg_passes -T1 -N1 -s3 -m 1 -t 1 -d " + collectionsDirectory + " -f " + collectionBaseName + " " + inputFiles)
	except Exception, data:
		s = "Error running mg_passes -T1 for MG collection build: " + `data` + "\n"
		writeDebugInfo(s)
		return FAILURE

	# Create *.text.dict
	try:
		os.system("mg_compression_dict -d " + collectionsDirectory + " -f " + collectionBaseName)
	except Exception, data:
		s = "Error running mg_compression_dict for MG collection build: " + `data` + "\n"
		writeDebugInfo(s)
		return FAILURE

	# Create *.invf.dict.hash
	try:
		os.system("mg_perf_hash_build -d " + collectionsDirectory + " -f " + collectionBaseName)
	except Exception, data:
		s = "Error running mg_perf_hash_build for MG collection build: " + `data` + "\n"
		writeDebugInfo(s)
		return FAILURE

	# Create *.text, *.text.idx, *.invf and *.invf.idx
	try:
		os.system("mg_passes -T2 -N2 -c 2 -t 1 -d " + collectionsDirectory + " -f " + collectionBaseName + " " + inputFiles)
	except Exception, data:
		s = "Error running mg_passes -T2 for MG collection build: " + `data` + "\n"
		writeDebugInfo(s)
		return FAILURE

	# Create *.text.idx.wgt and *.weight.approx
	try:
		os.system("mg_weights_build -d " + collectionsDirectory + " -f " + collectionBaseName + " -b 8")
	except Exception, data:
		s = "Error running mg_weights_build for MG collection build: " + `data` + "\n"
		writeDebugInfo(s)
		return FAILURE

	# Create *.invf.dict
	try:
		os.system("mg_invf_dict -d " + collectionsDirectory + " -f " + collectionBaseName + " -b 4096")
	except Exception, data:
		s = "Error running mg_invf_dict for MG collection build: " + `data` + "\n"
		writeDebugInfo(s)
		return FAILURE

	# Create *.text.dict
	try:
		os.system("mg_fast_comp_dict -d " + collectionsDirectory + " -f " + collectionBaseName)
	except Exception, data:
		s = "Error running mg_fast_comp_dict for MG collection build: " + `data` + "\n"
		writeDebugInfo(s)
		return FAILURE

	try:
		os.remove(inputFiles)
	except:
		pass

	#delete temporary collection files that are no longer needed
	permFileExtensions = ["TD", "T", "IDB", "TIW", "WA", "I"]
	userMgFiles = os.listdir(collectionsDirectory)
	lenCollectionName = len(collectionBaseName)
	for f in userMgFiles:
		if f[:lenCollectionName] == collectionBaseName:
			i = string.find(f, ".")
			if i != -1:
				if not string.upper(f[i+1:]) in permFileExtensions:
					try:
						os.remove(collectionsDirectory + f)
					except:
						pass

	return SUCCESS