#Module for calling MG (Managing Gigabytes) mgquery from a Python 1.5.2 Application
#running under NT 4.0.

import sys
import os
import string

def prepForBooleanDocnumsQuery(mgDir, userName, query):
	#use for the initial boolean query
	f = open(mgDir + userName + "/.mgrc", "w")
	f.write(".set mode docnums\n" +
			".set query boolean\n")
	f.close()
	return

def prepForDocnumsHeadsQuery(mgDir, userName, query):
	#use to get the document header (1st line) for the docnum'th document
	f = open(mgDir + userName + "/.mgrc", "w")
	f.write(".set mode heads\n" +
			".set query docnums\n" +
			".set heads_length 500\n")
	f.close()
	return

def prepForDocnumsTextQuery(mgDir, userName, query):
	#use to get the full text for the docnum'th document
	f = open(mgDir + userName  + "/.mgrc", "w")
	f.write(".set mode text\n" +
			".set query docnums\n")
	f.close()
	return

def readQueryResults(mgDir, userName):
	#read the results of the last query into a list and return
	f = open(mgDir + userName + "/qresults.txt")
	lines = f.readlines()
	f.close()
	return lines

def invokeMgQuery(mgDir, userName, docCollectionName):
	#invoke mgquery
	#
	#must use "<" and ">" instead of ".input" and ".output" commands
	#which cause a Dr. Watson dialog on my system.
	#
	os.system(mgDir + userName + "/mgquery" +
			  " -d " + mgDir + userName + "/" +
			  " -f " + docCollectionName +
			  " < " + mgDir + userName + "/query.txt" +
			  " > " +  mgDir + userName + "/qresults.txt")
	return

def runDocnumsQuery(mgDir, userName, docCollectionName, query):
	prepForBooleanDocnumsQuery(mgDir, userName, query)
	invokeMgQuery(mgDir, userName, docCollectionName)
	return

def runHeadsQuery(mgDir, userName, docCollectionName, query):
	prepForDocnumsHeadsQuery(mgDir, userName, query)
	invokeMgQuery(mgDir, userName, docCollectionName)
	return

def runTextQuery(mgDir, userName, docCollectionName, query):
	prepForDocnumsTextQuery(mgDir, userName, query)
	invokeMgQuery(mgDir, userName, docCollectionName)
	return

def writeQuery(mgDir, userName, query):
	#write the query to a file, which will take the place of STDIN
	#and keep mgquery from coming up in interactive mode.
	f = open(mgDir + userName + "/query.txt", "w")
	f.write(query + "\n")
	f.close()
	return

def mgRetrieve(mgDir, userName, docCollectionName, query):
	#retrieve the full text for the doc collection name and document number (query) passed
	os.chdir(mgDir + userName)
	writeQuery(mgDir, userName, query)
	runTextQuery(mgDir, userName, docCollectionName, query)
	lines = readQueryResults(mgDir, userName)
	text = ""
	for line in lines[2:]:
		text = text + line
	return text

def mgQuery(mgDir, userName, docCollectionName, query):
	#run query, returning results that consist of the document number followed by the
	#heading (1st line) for the document
	os.chdir(mgDir + userName)

	#run boolean query, getting document numbers for the result
	writeQuery(mgDir, userName, query)
	runDocnumsQuery(mgDir, userName, docCollectionName, query)
	results = readQueryResults(mgDir, userName)
	results = map(string.strip, results)

	#for each document number, get a heading (1st line of document)
	resultsForHtml = []
	for r in results[1:]:
		query = r
		writeQuery(mgDir, userName, query)
		runHeadsQuery(mgDir, userName, docCollectionName, query)
		headResults = readQueryResults(mgDir, userName)
		resultsForHtml.append(string.strip(headResults[0]))
	return resultsForHtml
