#handle html processing that provides proxying capability

import sys
import re
import string
import traceback
import socket
import time

ZAP_PORT = 8579		#port for communicating with ad-blocking process

origBaseUrl = ""

nonHtmlDocExtensions = ["aif", "aiff", "au", "avi", "bmp", "doc", "gif", "gtar", "gz", "gzip",
			   			"hqx", "jfif", "jpeg", "jpe", "jpg", "mp2", "mp3",
					   	"mov", "movie", "mpeg", "mpg", "mpe", "png",
			   			"pdf", "ps", "qt", "ram", "rm", "ra", "sit", "snd",
			   			"tar", "tgz", "tiff", "tif", "txt", "wav", "xbm", "xls", "Z", "zip"]

htmlDocExtensions = ["asp", "htm", "html", "jsp"]

javascriptEvents = ["onclick", "oncontextmenu", "ondblclick", "ondrag", "ondragend",
					"ondragenter", "ondragleave", "ondragover", "ondragstart", "ondrop",
					"onhelp", "onkeydown", "onkeypress", "onkeyup", "onmousedown",
					"onmousemove", "onmouseout", "onmouseover", "onmouseup",
					"onpropertychange", "onreadystatechange"]

def writeDebugInfo(msg):
	debugFile = open("c:/debug.txt", "a+")
	debugFile.write(msg)
	debugFile.close()
	return

def tracebackAsString(limit):
	type, value, tb = sys.exc_info()
	aList = traceback.format_exception(type, value, tb, limit)
	formattedString = ""
	for s in aList:
		formattedString = formattedString + s + "\n"
	return formattedString

def convertUrlIfForAdvertisement(absoluteUrl):
	#convert advertisement URLs to invisible gif URLs

	#use sockets to talk with separate process which will handle ad-blocking
	try:
		clientsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		clientsock.connect("localhost", ZAP_PORT)
		clientsock.send("zapcheck|" + absoluteUrl)
		result = clientsock.recv(1024)
		clientsock.close()
	except Exception, data:
		result = "error"
		writeDebugInfo("Error communicating with ad-blocking process: " + `data` + "\n")
		try:
			clientsock.close()
		except:
			pass

	#writeDebugInfo("zapAds result=" + result + "\n")

	if result == "block":
		writeDebugInfo("BLOCKED " + absoluteUrl + "\n")
		absoluteUrl = "/proxyDemo/bin/dot_clear.gif"

	return absoluteUrl

def getBaseUrlFromFullUrl(fullUrl):
	global origBaseUrl

	#get rid of http://
	if fullUrl[:7] == "http://":
		fullUrl = fullUrl[7:]

	#get rid of query string
	i = string.find(fullUrl, "?")
	if i != -1:
		fullUrl = fullUrl[:i]

	#chop off end part if it refers to a document versus a folder
	i = string.rfind(fullUrl, "/")
	j = string.rfind(fullUrl, ".")

	if i != -1:
		if j > i:
			fullUrl = fullUrl[:i+1]

	origBaseUrl = fullUrl[:]

	return origBaseUrl

def constructAbsoluteUrl(url):
	global origBaseUrl

	#tag is absolute
	if url[:7] == "http://":
		absoluteUrl = url

	elif url[:4] == "www.":
		absoluteUrl = "http://" + url

	#fix relative tag beginning with / which may change the origBaseUrl as well
	elif url[0] == "/":
		tmpOrigBaseUrl = origBaseUrl[:]
		i = string.find(tmpOrigBaseUrl, "/")
		if i != -1:
			tmpOrigBaseUrl = tmpOrigBaseUrl[:i]
		absoluteUrl = "http://" + tmpOrigBaseUrl + url

	#fix relative url beginning with ../  etc.
	elif url[:3] == "../":
		tmpOrigBaseUrl = origBaseUrl[:]
		while url[:3] == "../":
			url = url[3:]
			if tmpOrigBaseUrl and tmpOrigBaseUrl[-1] == "/":
				tmpOrigBaseUrl = tmpOrigBaseUrl[:-1]
			i = string.rfind(tmpOrigBaseUrl, "/")
			if i != -1:
				tmpOrigBaseUrl = tmpOrigBaseUrl[:i+1]
		absoluteUrl = "http://" + tmpOrigBaseUrl + url

	#add http:// to tag beginning with origBaseUrl
	elif url[:len(origBaseUrl)] == origBaseUrl:
		absoluteUrl = "http://" + url

	#add http:// and origBaseUrl to relative tag which doesn't begin with a special character
	else:
		if origBaseUrl[-1] == "/":
			tmpOrigBaseUrl = origBaseUrl[:-1]
		else:
			tmpOrigBaseUrl = origBaseUrl
		absoluteUrl = "http://" + tmpOrigBaseUrl + "/" + url

	return absoluteUrl

def anchorForNonHtmlDoc(url):
	#see if extension of doc is in "special" list of non-HTML docs
	urlParts = string.split(url, ".")
	if len(urlParts) > 1:
		if string.lower(urlParts[-1]) in nonHtmlDocExtensions:
			return 1
	return 0

def anchorReplaceFunction(match):
	global origBaseUrl
	try:
		anchorTagBegin = match.group("anchorTagBegin")
		anchorUrl = match.group("anchorUrl")
		#check for internal (same page) anchor--don't mess with them
		if anchorUrl[0] == "#":
			s = anchorTagBegin + anchorUrl
		#check for "mailto:" anchor--don't mess with them either
		elif anchorUrl[:7] == "mailto:":
			s = anchorTagBegin + anchorUrl
		#check for anchor points to a target which we can't modify, such as a .PDF document
		elif anchorForNonHtmlDoc(anchorUrl):
			if anchorUrl[:7] == "http://":
				s = anchorTagBegin + anchorUrl
			else:
				s = anchorTagBegin + "http://" + origBaseUrl + anchorUrl
		else:
			anchorUrl = string.replace(anchorUrl, "?", "&")
			s = anchorTagBegin + '/proxyDemo/bin/proxy?k0olbase=' + origBaseUrl + '&k0olurl=' + anchorUrl
		return s
	except Exception, data:
		raise Exception, "Error in anchorReplaceFunction: anchorTagBegin=" + anchorTagBegin + " anchorUrl=" + anchorUrl

def metaRefreshReplaceFunction(match):
	metaTagBegin = match.group("metaTagBegin")
	metaRefreshUrl = match.group("metaRefreshUrl")
	metaRefreshUrl = string.replace(metaRefreshUrl, "?", "&")
	s = metaTagBegin + '/proxyDemo/bin/proxy?k0olbase=' + origBaseUrl + '&k0olurl=' + metaRefreshUrl
	return s

def bgrndUrlReplaceFunction(match):
	bodyTagBegin = match.group("bodyTagBegin")
	bgrndUrl = match.group("bgrndUrl")
	absoluteUrl = constructAbsoluteUrl(bgrndUrl)
	absoluteUrl = convertUrlIfForAdvertisement(absoluteUrl)
	s = bodyTagBegin + absoluteUrl
	return s

def srcReplaceFunction(match):
	srcTagBegin = match.group("srcTagBegin")
	srcUrl = match.group("srcUrl")
	#frame tag src attribute--handle differently
	if string.lower(srcTagBegin[:2]) == "<f":
		s = srcTagBegin + '/proxyDemo/bin/proxy?k0olbase=' + origBaseUrl + '&k0olurl=' + srcUrl
	#non-frame tag src attribute
	else:
		absoluteUrl = constructAbsoluteUrl(srcUrl)
		absoluteUrl = convertUrlIfForAdvertisement(absoluteUrl)
		s = srcTagBegin + absoluteUrl
	return s

def formActionUrlReplaceFunction(match):
	#to pass OUR query string parms when a form action is invoked, we must make
	#them hidden form fields.
	global origBaseUrl
	formTagBegin = match.group("formTagBegin")
	actionUrl = match.group("actionUrl")
	rest = match.group("rest")
	formStuff = match.group("formStuff")
	formEnd = match.group("formEnd")

	#for debugging
	#writeDebugInfo("formTagBegin is " + formTagBegin + "\n"
	#				+ "actionUrl is " + actionUrl + "\n"
	#				+ "rest is " + rest + "\n\n")

	s = (formTagBegin + "/proxyDemo/bin/proxy" + rest
		+ formStuff + "\n"
		+ '<input type=hidden name=k0olbase value="' + origBaseUrl + '">' + "\n"
		+ '<input type=hidden name=k0olurl value="' + actionUrl + '">' + "\n"
		+ formEnd)
	return s

def jsEventCodeReplaceFunction(match):
	s = match.group("stringContents")
	s = string.strip(s)
	i = string.rfind(s, ".")
	if i != -1 and len(s) > i:
		lastPart = string.lower(s[i+1:])
		if (lastPart in nonHtmlDocExtensions) or (lastPart in htmlDocExtensions):
			s = constructAbsoluteUrl(s)
			s = convertUrlIfForAdvertisement(s)
	return "'" + s + "'"

def jsEventReplaceFunction(match):
	global javascriptEvents
	jsEventTagBegin = match.group("jsEventTagBegin")
	jsEventCode = match.group("jsEventCode")

	#get event name
	jsEvent = string.split(jsEventTagBegin, "=")
	jsEvent = jsEvent[0]
	jsEvent = string.strip(jsEvent)

	#writeDebugInfo("jsEvent is " + jsEvent + " and jsEventCode is " + jsEventCode + "\n")

	if string.lower(jsEvent) in javascriptEvents:
		stringPat = re.compile(r'\'(?P<stringContents>.*?)\'',
							   re.IGNORECASE | re.DOTALL)
		jsEventCode = stringPat.sub(jsEventCodeReplaceFunction, jsEventCode)

	s = jsEventTagBegin + jsEventCode

	return s

def extScriptUrlReplaceFunction(match):
	#extScriptPatBegin = match.group("extScriptPatBegin")
	#extScriptUrl = match.group("extScriptUrl")
	#absoluteUrl = constructAbsoluteUrl(extScriptUrl)
	#s = extScriptPatBegin + absoluteUrl

	s = ""	#remove external scripts for now
	return s

def proxifyHtml(s):
	#fix anchor urls
	anchorPat = re.compile(r'(?P<anchorTagBegin><[^>]+?\b(href|usemap)\s*?=\s*?["\']?)(?P<anchorUrl>[^"\' >]+)',
						   re.IGNORECASE | re.DOTALL)
	s = anchorPat.sub(anchorReplaceFunction, s)

	#fix meta refresh url
	metaRefreshPat = re.compile(r'(?P<metaTagBegin><meta[^>]+?\burl\s*?=\s*?)(?P<metaRefreshUrl>[^"\' >]+)',
						   re.IGNORECASE | re.DOTALL)
	s = metaRefreshPat.sub(metaRefreshReplaceFunction, s)

	#find external script tags--do this before the other tags with "src" in them
	extScriptPat = re.compile(r'(?P<extScriptPatBegin><script[^>]+?\bsrc\s*?=\s*?["\']?)(?P<extScriptUrl>[^<]*?</script>)',
						 	  re.IGNORECASE | re.DOTALL)
	s = extScriptPat.sub(extScriptUrlReplaceFunction, s)

	#fix src tag urls--will potentially have to do something different when we start keeping script tags
	#because these have "src" attributes as well.
	srcPat = re.compile(r'(?P<srcTagBegin><[^>]+?[^.]\b(low|dyn)?src\s*?=\s*?["\']?)(?P<srcUrl>[^"\' >]+)',
						   re.IGNORECASE | re.DOTALL)
	s = srcPat.sub(srcReplaceFunction, s)

	#fix image objects' (dot) src properties (Javascript).
	#NOTE: shares replace function (srcReplaceFunction) with src tag urls
	imageObjectSrcPropertyPat = re.compile(r'(?P<srcTagBegin>\.src\s*?=\s*?["\'])(?P<srcUrl>[^"\' >]+)',
						   				   re.IGNORECASE | re.DOTALL)
	s = imageObjectSrcPropertyPat.sub(srcReplaceFunction, s)

	#fix events (Javascript)--all events starting with "on"--see global list
	javascriptEventPat = re.compile(r'(?P<jsEventTagBegin> on[a-zA-Z]+?\s*?=\s*?")(?P<jsEventCode>[^"]+)',
									re.IGNORECASE | re.DOTALL)
	s = javascriptEventPat.sub(jsEventReplaceFunction, s)

	#fix body tag background url
	bodyBgrndPat = re.compile(r'(?P<bodyTagBegin><body[^>]+?\bbackground\s*?=\s*?["\']?)(?P<bgrndUrl>[^"\' >]+)',
							  re.IGNORECASE | re.DOTALL)
	s = bodyBgrndPat.sub(bgrndUrlReplaceFunction, s)

	#fix form action urls
	formActionPat = re.compile(r'(?P<formTagBegin><form[^>]+?\baction\s*?=\s*?["\']?)(?P<actionUrl>[^"\' >]+)(?P<rest>.*?>)(?P<formStuff>.*?)(?P<formEnd></form>)',
						 	   re.IGNORECASE | re.DOTALL)
	s = formActionPat.sub(formActionUrlReplaceFunction, s)

	return s