#This is a server which talks to Zope via sockets. It takes requests from a
#Zope external method to store and index html documents, including full-text indexing.

#To Do:
#1) Eventually, files to index should be given a special extension so that this program will find them even
#	if something goes wrong with the communications from Zope. Eventually, while not busy, this program
#	should check for files with the special extension in the temp dir and should process then delete them.
#2)	Fix sys.path.append statement in this file (storagerIndexer.py) to not be hardcoded to the Zope
#	installation path on a particular user's machine
#3)	boolean searching with parens (and maybe otherwise) not working currently; regular searches work
#	fine (or seem to)

#Done:
#1) save original html as a separate compressed file under user name folder
#2) stemming using Porter.py
#3)	strip stopwords using words in english.stop and stop_words.txt files
#4) handle searching also
#5) reorg storagerIndexer to make it recognize difference between a storage request and
#	a search request
#	pseudocode:
#		a) proxy sends search request, which has two args: "search" and the contents of the query
#		b) this program does the query and returns the results as proxified html via sockets mech
#
#	add dateAndTime arg and field and index and column so we can save and retrieve different versions
#	of a site (saved at different dates/times)
#6) strip out all html before saving text (in ZODB) for indexing purposes
#7) add back sleep statement in loop to give other programs more time to run--frozen screensaver syndrome
#8) add back content as a "column" since the search functionality will have to return it
#9) move process html stuff to a separate module which can be used by this program and by the proxy
#10)eliminate all columns except "fields" which has all the info. we need separated by "|" chars, so that
#	we need to parse out the pieces of info. when we need them
#11)eliminate all indices except the "content" text index; store the UID in the content and search for
#	it there

#In process:

import sys

sys.path.append("h:/website/lib/python")

import ZODB
from ZODB.FileStorage import FileStorage

import Persistence
from Persistence import Persistent

from Products.ZCatalog.Catalog import Catalog
from Products.ZCatalog import Vocabulary

import traceback
import socket
import select
import string
import os
import time
import gzip
import tempfile
import re

import Porter

import processHtml
import removeHtml
import mgQuery
import buildCollection

HOST = ''			#this means local
PORT = 8578			#arbitrary, high number

def tracebackAsString(limit):
	type, value, tb = sys.exc_info()
	aList = traceback.format_exception(type, value, tb, limit)
	formattedString = ""
	for s in aList:
		formattedString = formattedString + s + "\n"
	return formattedString

def writeDebugInfo(msg):
	debugFile = open("c:/debug.txt", "a+")
	debugFile.write(msg)
	debugFile.close()
	return

def decompressSurfedFile(compressedFileName):
	#writeDebugInfo("decompressing " + compressedFileName + "\n")
	#f = gzip.GzipFile(compressedFileName)
	f = open(compressedFileName)	#not really compressed
	lines = f.readlines()
	f.close()
	s = ""
	for line in lines:
		s = s + line
	return s

def createStorageFoldersIfNecessary(storageFolder, userName):
	#create folder for storing user ZODBs if necessary
	if not os.path.exists(storageFolder):
		#writeDebugInfo("Storage folder for surfed sites does not exist! Creating...\n")
		try:
			os.makedirs(storageFolder)
			#writeDebugInfo("Successfully created storage folder.\n")
		except:
			writeDebugInfo("Error creating storage folder.\n")
			return

	#create folder for storing compressed pages for user if necessary
	if not os.path.exists(storageFolder + "\\" + userName):
		#writeDebugInfo("Storage folder for user '" + userName
		#				+ "' does not exist! Creating...\n")
		try:
			os.makedirs(storageFolder + "\\" + userName)
			#writeDebugInfo("Successfully created storage folder for user '" + userName + "'.\n")
		except:
			writeDebugInfo("Error creating storage folder for user '" + userName + "'.\n")
			return
	return

def storeSurfedFileCompressed(fileName, storageFolder, userName, siteUrl):
	#NOT  using compression right now--were having a problem with decompressing, so have
	#commented out all compressing and decompressing for now.
	#
	#read the temp file
	f = open(fileName)
	lines = f.readlines()
	f.close()

	siteHtmlString = ""
	for line in lines:
		siteHtmlString = siteHtmlString + line

	#add some meta info. to the file before storing it--needs to be where MG can return
	#it (the meta info) as part of a header query
	headBeginTagPat = re.compile(r'<html>\s*?<head>', re.IGNORECASE | re.DOTALL)
	match = headBeginTagPat.search(siteHtmlString)

	#if no <html><head> pattern, we will wrap what's there
	if not match:
		htmlWithMetaData = ("<html><head>"
							+ '<meta name="origK0olUrl" content="' + siteUrl + '">'
							+ '<meta name="k0olTime" content="' + time.ctime(time.time()) + '">\n'
							+ siteHtmlString + "<\head><\html>")
	else:
		endHeadBegin = match.end()
		htmlWithMetaData = (siteHtmlString[:endHeadBegin]
							+ '<meta name="origK0olUrl" content="' + siteUrl + '">'
							+ '<meta name="k0olTime" content="' + time.ctime(time.time()) + '">\n'
							+ siteHtmlString[endHeadBegin+1:])

	#get a temp file name in user's surfed pages folder
	tempfile.tempdir = storageFolder + "\\" + userName
	fn = tempfile.mktemp()
	fn = string.replace(fn, "~", userName)
	#fn = fn + ".gz"
	fn = fn + ".html"	#make a regular file versus a gzipped file

	#create a compressed file, write data to it, etc.
	#cf = gzip.GzipFile(fn, "wb", 9)

	#just create a regular file for now (versus compressed)
	cf = open(fn, "w")
	cf.write(htmlWithMetaData)
	cf.close()

	#return the file name
	return fn

def getFieldFromSearchResult(searchResult, fieldName):
	fields = searchResult["fields"]
	siteUrl, dateAndTime, compressedFileName, pageId = string.split(fields, "|")
	if fieldName == "siteUrl":
		field = siteUrl
	elif fieldName == "dateAndTime":
		field = dateAndTime
	elif fieldName == "compressedFileName":
		field = compressedFileName
	elif fieldName == "pageId":
		field = pageId
	else:
		field = None
	return field

def getSiteUrlFromSearchResult(searchResult):
	urlPat = re.compile(r'<meta name="origK0olUrl" content="', re.IGNORECASE | re.DOTALL)
	match = urlPat.search(searchResult)
	siteUrl = ""
	if match:
		urlBegin = match.end()
		#writeDebugInfo("searchResult=" + searchResult + "\n")
		for c in searchResult[urlBegin:]:
			if c == '"':
				break
			else:
				siteUrl = siteUrl + c
	return siteUrl

def getDateTimeFromSearchResult(searchResult):
	dateTimePat = re.compile(r'<meta name="k0olTime" content="', re.IGNORECASE | re.DOTALL)
	match = dateTimePat.search(searchResult)
	dateTime = ""
	if match:
		dateTimeBegin = match.end()
		#writeDebugInfo("searchResult=" + searchResult + "\n")
		for c in searchResult[dateTimeBegin:]:
			if c == '"':
				break
			else:
				dateTime = dateTime + c
	return dateTime

class SurfedFile(Persistent):
	def __init__(self, siteUrl, dateAndTime, fileName, compressedFileName):
		self.fields = siteUrl + "|" + dateAndTime + "|" + compressedFileName + "|"
		#self.siteUrl = siteUrl
		#self.dateAndTime = dateAndTime
		#self.compressedFileName = compressedFileName
		f = open(fileName)
		lines = f.readlines()
		f.close()
		s = ""

		for line in lines:
			s = s + line

		s = removeHtml.stripHtml(s)	#strip all HTML from version to be indexed

		#add special control words used for searching instead of having more indices;
		#these probably aren't going to work because full-text indices appear to only work
		#on words--not including numbers as separate items like in dates, etc.
		#self.content = s + " @@" + siteUrl + "@@ @@" + dateAndTime + "@@ "

		self.content = s
		return

class SurfedDb:
	def __init__(self, storageFolder, userName):
		self.file = storageFolder + "/" + userName + "DB.fs"
		self.db = ZODB.DB(FileStorage(self.file))
		self.co = self.db.open()
		self.root = self.co.root()

		#if we opened an existing ZODB with a catalog,
		#point the SurfedDb's cat (catalog) attribute to the ZODB's cat (catalog)
		if self.root.has_key("cat"):
			self.cat = self.root["cat"]

		#otherwise, this is a new ZODB; so, create and init the catalog and set a "pointer" to it
		else:
			#full-text indices need vocabularies
			v = Vocabulary.Vocabulary('Vocabulary', 'Vocabulary', globbing=0)
			self.cat = Catalog(v)
			self.cat.aq_parent = self.root
			self.root['cat'] = self.cat

			#columns are needed for whatever you want to report on
			#self.cat.addColumn("siteUrl")
			#self.cat.addColumn("dateAndTime")
			#self.cat.addColumn("compressedFileName")
			#self.cat.addColumn("pageId")
			#self.cat.addColumn("content")
			self.cat.addColumn("fields")

			#indices are needed for whatever you want to search on;
			#will have to add these back as field indices if we ever want to search
			#on them; we can't search on them as part of the text content because they
			#consist of multiple words (including numbers potentially) and text indices
			#only index words (things starting with alphas). We could, I guess, play tricks
			#by squishing special chars etc out of the data before storing it and doing the
			#same to the user's query info. Worry about this later.
			#self.cat.addIndex("siteUrl", "FieldIndex")
			#self.cat.addIndex("dateAndTime", "FieldIndex")

			#text index only works on words, meaning, something that starts with an alpha
			#and doesn't contain any special characters; numbers such as 1234 aren't indexed,
			#but a word such as number1234 is indexed.
			self.cat.addIndex("content", "TextIndex")

			get_transaction().commit()
		return

	def close(self):
		self.db.close()
		return

	def storeSurfedFile(self, surfedFile):
		uid = id(surfedFile)
		surfedFile.fields = surfedFile.fields + `uid`
		surfedFile.content = surfedFile.content + " uid" + `uid` + " "
		self.root[uid] = surfedFile
		self.cat.catalogObject(surfedFile, uid)
		get_transaction().commit()
		return

def serve():
	sdb = None
	serversock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	serversock.bind(HOST, PORT)
	serversock.listen(5)

	print "StoragerIndexer is ready for requests...\n"

	try:
		while 1:
			#give other programs some time to run
			time.sleep(0.1)

			#poll the socket for activity
			input, output, exc = select.select([serversock], [], [], 0)

			for sock in input:
				try:
					handlersock, addr = sock.accept()
				except Exception, data:
					print "Error accepting client socket: " + `data` + "\n"
					print tracebackAsString(25)
					break

				try:
					rcvdata = handlersock.recv(1024)
				except Exception, data:
					try:
						handlersock.close()
					except:
						pass
					print "Error receiving client data: " + `data` + "\n"
					print tracebackAsString(25)
					break

				try:
					argsList = string.split(rcvdata, "|")
					command = argsList[0]
					if command == "store":
						handlersock.close()
						userName, fileName, storageFolder, siteUrl, dateAndTime = argsList[1:]
						#print "username=" + `userName` + "\n"
						#print "filename=" + `fileName` + "\n"
						#print "storagefolder=" + `storageFolder` + "\n"
						#print "siteurl=" + `siteUrl` + "\n"
						#print "dateAndTime=" + `dateAndTime` + "\n"
					elif command == "query":
						userName, storageFolder, query, mgDir, mgCollection = argsList[1:]
						#print "userName=" + `userName` + "\n"
						#print "storageFolder=" + `storageFolder` + "\n"
						#print "query=" + `query` + "\n"
						#print "mgDir=" + `mgDir` + "\n"
						#print "mgCollection=" + `mgCollection` + "\n"
					elif command == "retrieve":
						userName, storageFolder, pageId, mgDir, mgCollection = argsList[1:]
					elif command == "build":
						userName, tempDirectory, storageFolder, mgDir, mgCollection = argsList[1:]
					else:
						pass	#handle shutdown command eventually

					#Close current ZODB if a different user this time
					if sdb != None:
						if userName != lastUserName:
							sdb.close()
							sdb = None

					#open ZODB for user if necessary
					if sdb == None:
						sdb = SurfedDb(storageFolder, userName)

					lastUserName = userName

					if command == "store":
						createStorageFoldersIfNecessary(storageFolder, userName)
						compressedFileName = storeSurfedFileCompressed(fileName, storageFolder, userName, siteUrl)
						sf = SurfedFile(siteUrl, dateAndTime, fileName, compressedFileName)
						sdb.storeSurfedFile(sf)
						os.remove(fileName)
					elif command == "query":
						originalQuery = query
						#writeDebugInfo("query=" + `query` + "\n")

						#first stem the query since the text has been stemmed before storage in ZODB
						queryList = string.split(query, " ")

						#hadEndingParen = []

						#for i in range(0, len(queryList)):
						#	if string.find(queryList[i], ")") != -1:
						#		hadEndingParen.append(1)
						#	else:
						#		hadEndingParen.append(0)

						queryList = Porter.stem(queryList)

						query = ""
						for i in range(0, len(queryList)):
							#if hadEndingParen[i]:
							#	if string.find(queryList[i], ")") == -1:
							#		query = query + queryList[i] + ")" + " "
							#	else:
							#		query = query + queryList[i] + " "
							#else:
							query = query + queryList[i] + " "

						query = string.strip(query)

						#writeDebugInfo("query after stemming =" + `query` + "\n")

						#now perform the query
						results = sdb.cat({"content" : query})

						#perform a query against MG collection as well
						mgResults = mgQuery.mgQuery(mgDir, userName, mgCollection, originalQuery)

						s = ("<html>" + "\n"
							+ "\t<head>" + "\n"
							+ "\t</head>" + "\n"
							+ "\t<body>" + "\n"
							+ "<h1>Search Results</h1>" + "<br>\n\n")

						if results or mgResults:
							for result in results:
								s = s + ('\t\t<a href="getPage?pageId=' + getFieldFromSearchResult(result,"pageId") + '">'
										+ '<font size="5">' + getFieldFromSearchResult(result,"siteUrl") + '</font>'
										+ '</a>'
										+ " at "
										+ getFieldFromSearchResult(result,"dateAndTime") + "<br><br>\n")
							for result in mgResults:
								i = string.find(result, " ")
								if i == -1:
									continue
								docId = "mg" + result[:i]
								urlFromDoc = getSiteUrlFromSearchResult(result)
								dateTimeFromDoc = getDateTimeFromSearchResult(result)

								s = s + ('\t\t<a href="getPage?pageId=' + docId + '">'
										+ '<font size="5">' + urlFromDoc + '</font>'
										+ '</a>'
										+ " at "
										+ dateTimeFromDoc + "<br><br>\n")
						else:
							s = s + '<font size="5">No matches found.</font><br><br>\n'

						s = s + ("\t</body>" + "\n"
								+ "</html>" + "\n")

						handlersock.send(s)
						handlersock.close()
					elif command == "retrieve":
						if pageId[:2] == "mg":
							query = pageId[2:]
							mgResults = mgQuery.mgRetrieve(mgDir, userName,
															mgCollection, query)
							processHtml.origBaseUrl = getSiteUrlFromSearchResult(mgResults)
							s = processHtml.proxifyHtml(mgResults)
							handlersock.send(s)
							handlersock.close()
						else:
							results = sdb.cat({"content" : "uid" + pageId})

							#writeDebugInfo("\n\n---------retrieve results=\n\n" + `results` + "\n\n----------\n\n")

							compressedFileName = getFieldFromSearchResult(results[0],"compressedFileName")
							processHtml.origBaseUrl = getFieldFromSearchResult(results[0],"siteUrl")
							s = decompressSurfedFile(compressedFileName)
							s = processHtml.proxifyHtml(s)

							handlersock.send(s)
							handlersock.close()
					elif command == "build":
						if (buildCollection.buildCollection(userName, tempDirectory,
															storageFolder, mgDir, mgCollection)
							== buildCollection.SUCCESS):
							#delete the user's Zope daily surfed files database since the files
							#are now in the MG collection
							if sdb != None:
								try:
									sdb.close()
								except:
									pass
								sdb = None
							zodbFileNameSuffixes = ["DB.fs","DB.fs.index","DB.fs.lock","DB.fs.tmp"]
							for fn in zodbFileNameSuffixes:
								try:
									fullfn = storageFolder + "/" + userName + fn
									os.remove(fullfn)
								except Exception, data:
									writeDebugInfo("Error deleting daily surfed files archive file "
													+ fullfn + "\n" + `data` + "\n")
					else:
						handlersock.close()

				except Exception, data:
					print "Error processing client request: " + `data` + "\n"
					print tracebackAsString(25)
					try:
						sdb.close()
						sdb = None
					except:
						pass
					try:
						handlersock.close()
					except:
						pass
					break

	except Exception, data:
		print "Error: " + `data`
		print tracebackAsString(25)
		try:
			serversock.close()
		except:
			pass
		try:
			sdb.close()
		except:
			pass

	return

#if __name__ == "__main__":
#	serve()
serve()